#
# $Id: Lojban.pm,v 1.5 1998/11/25 00:57:14 cvs Exp $
#

package Lingua::Lojban;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);

require Exporter;

@ISA = qw(Exporter AutoLoader);
# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.
@EXPORT = ();
@EXPORT_OK = qw(
  gismu 
  rafsilist rafsi2gismu rafsi4gismu make_lujvo gismulist split_lujvo
);
$VERSION = '0.02';


# Preloaded methods go here.

use GDBM_File;
use Storable qw(freeze thaw);

# find the database files. Need to handle both pre-install and post-install.
# the files should go near Lojban.pm, but should be in an arch-specific
# directory since gdbm files are endian-dependent.

my($gismu_gdbm,$rafsi_gdbm);
foreach (@INC) {
    if (-r "$_/gismu.gdbm") {
	$gismu_gdbm = "$_/gismu.gdbm";
	last;
    }
}
foreach (@INC) {
    if (-r "$_/rafsi.gdbm") {
	$rafsi_gdbm = "$_/rafsi.gdbm";
	last;
    }
}
die "couldn't find gismu database in \@INC" unless $gismu_gdbm;
die "couldn't find rafsi database in \@INC" unless $rafsi_gdbm;

my(%rawgismu, %gismu_cache);

# init
tie %rawgismu, "GDBM_File", $gismu_gdbm, &GDBM_READER, 0644
  or die "couldn't read gismu database $gismu_gdbm: $!";

my(%rawrafsi, %rafsi);
tie %rawrafsi, "GDBM_File", $rafsi_gdbm, &GDBM_READER, 0644
  or die "couldn't read rafsi database $rafsi_gdbm: $!";

my($consonants) = 'bcdfgjklmnprstvxz';
my(@consonants) = split(//,$consonants);
my(@unvoiced_consonants) = split(//,'ptkfcsx');
my(@voiced_consonants) =   split(//,'bdgvjz');
my(@init_cpairs) = qw(
		      bl br
		      cf ck cl cm cn cp cr ct
		      dj dr dz
		      fl fr
		      gl gr
		      jb jd jg jm jv
		      kl kr
		      ml mr
		      pl pr
		      sf sk sl sm sn sp sr st
		      tc tr ts
		      vl vr
		      xl xr
		      zb zd zg zm zv
); # refgrammar, p. 37
my(@cpairs);
{ # refgrammar, p. 36
    foreach my $a (@consonants) {
	foreach my $b (@consonants) {
	    # rule 1: no double consonants
	    next if $a eq $b;
	    # rule 2: one consonant voiced, other unvoiced is forbidden
	    next if (grep {$a eq $_} @unvoiced_consonants and
		     grep {$b eq $_} @voiced_consonants) or
		       (grep {$b eq $_} @unvoiced_consonants and
			grep {$a eq $_} @voiced_consonants);
	    # rule 3: both consonants may not be from the set 'cjsz'
	    next if grep {$a eq $_} qw(c j s z) and
	      grep {$b eq $_} qw(c j s z);
	    # rule 4: specific forbidded pairs
	    next if grep {"$a$b" eq $_} qw(cx kx xc xk mz);
	    push(@cpairs, "$a$b");
	}
    }
}
#print "found ",scalar(@cpairs)," consonant pairs\n";

my(@forbidden_ctriples) = qw(ndj ndz ntc nts);

sub valid_consonant_cluster {
    my($cluster, $isname) = @_;
    unless ($isname) {
	if (length($cluster) == 2) {
	    return 1 if grep {$cluster eq $_} @cpairs;
	    return 0;
	}
	if (length($cluster) == 3) {
	    return 1 if grep {substr($cluster, 0, 2) eq $_} @init_cpairs and
	      grep {substr($cluster, 1, 2) eq $_} @cpairs and
		!grep {$cluster eq $_} @forbidden_ctriples;
	    return 0;
	}
    }
    # names may have longer clusters, may start with any permissible pair, and
    # don't have forbidden triples
    foreach my $i (0 .. length($cluster)-2) {
	return 0 unless grep {substr($cluster, $i, 2) eq $_} @cpairs;
    }
    return 1;
}

# returns a hash that describes everything there is to know about this gismu
# gismu, rafsi1, rafsi2, rafsi3, english, metaphor, place_structure, code,
# number, refs
sub gismu {
    my($gismu) = @_;
    return $gismu_cache{$gismu} if $gismu_cache{$gismu};
    # must thaw it out
    my $frozen = $rawgismu{$gismu};
    #die "$gismu is undefined" unless $frozen;
    if (defined($frozen)) {
	$gismu_cache{$gismu} = thaw($frozen);
    } else {
	$gismu_cache{$gismu} = undef;
    }
    return $gismu_cache{$gismu};
}

# returns an array of all the gismu, in random order
my(@gismulist_cache);
sub gismulist {
    @gismulist_cache = keys(%rawgismu) unless @gismulist_cache;
    return(@gismulist_cache);
}

# returns an array of all the rafsi, in random order
my(@rafsilist_cache);
sub rafsilist {
    @rafsilist_cache = keys(%rawrafsi) unless @rafsilist_cache;
    return(@rafsilist_cache);
}

my(%short_rafsi);
# returns the list of short rafsi associated with the gismu, if any
sub rafsi4gismu {
    my($gismu) = @_;
    return(@{$short_rafsi{$gismu}}) if $short_rafsi{$gismu};
    my $g = gismu($gismu);
    my(@rafsi) = grep {$_} (@$g{'rafsi1','rafsi2','rafsi3'});
    $short_rafsi{$gismu} = \@rafsi;
    return @rafsi;
}

# returns the gismu that this rafsi represents
sub rafsi2gismu {
    my($rafsi) = @_;
    return $rafsi{$rafsi} if $rafsi{$rafsi};
    $rafsi{$rafsi} = $rawrafsi{$rafsi};
    return $rafsi{$rafsi};
}

# categorize the word into [CVy], ignoring apostrophes
sub categorize {
    my($form) = @_;
    $form =~ s/\'//g;
    $form =~ s/[aeiou]/V/g;
    $form =~ s/[$consonants]/C/g;
    die unless $form =~ /^[CVy]*$/;
    return $form;
}

my $make_lujvo_verbose = 0;

sub make_lujvo {
    my(@gismu) = @_;
    die "go away, we're, like, closed or something" unless @gismu > 1;
    # initial/middle gismu can use the 4-letter rafsi, the end can use 5-letter
    my(@rafsis);
    foreach (0 .. $#gismu-1) {
	push(@{$rafsis[$_]}, 
	     substr($gismu[$_], 0, 4), rafsi4gismu($gismu[$_]));
    }
    push(@{$rafsis[$#gismu]}, 
	 $gismu[$#gismu], rafsi4gismu($gismu[$#gismu]));
    if ($make_lujvo_verbose) {
	foreach (0 .. $#rafsis) {
	    print " gismu ",$gismu[$_],": ",join(",",@{$rafsis[$_]}),"\n";
	}
    }
    # test this against p. 71
    # now must recurse to build up a list of possible lujvo
    my(@potential) = make_lujvo_recurse(\@rafsis, 0);
    print " potential: ",join(',', map { join('-',@$_) } @potential),"\n"
      if $make_lujvo_verbose;
    my(@lujvo,%scores);
    foreach my $potential (@potential) {
	my(@l) = @$potential;
	print "  testing (",join('-',@l),")\n"
	  if $make_lujvo_verbose;
	# categorize the pieces into CVCCVVCCVC... forms
	my(@f) = map { categorize($_) } @l;
	print "   form ",join('-',@f),"\n"
	  if $make_lujvo_verbose;
	unless ($l[$#l] =~ /[aeiou]$/) {
	    # must end in a vowel
	    print "   doesn't end in a vowel\n"
	      if $make_lujvo_verbose;
	    next;
	}
	my $score_h = 0;
	my $score_r = 0;
	# do the first 5 letters have a consonant cluster?
	#my $a = join('', @l);
	#$a =~ s/\'//g;
	#unless (substr($a,0,5) =~ /[$consonants]{2}/) {
	#    print "   doesn't have a consonant cluster in the 1st 5 letters\n";
	#    next;
	#}
	# y-hyphen is used after a CVC rafsi when the following rafsi would
	# cause an impermissible consonant pair, or when the result could
	# break up
	foreach (0 .. $#l-1) {
	    my $f = $f[$_];
	    # scoring
	    $score_r += 1 if $f eq "CVCCV";
	    $score_r += 2 if $f eq "CVCC";
	    $score_r += 3 if $f eq "CCVCV";
	    $score_r += 4 if $f eq "CCVC";
	    $score_r += 5 if $f eq "CVC";
	    $score_r += 6 if $f eq "CVV" and length($l[$_]) == 4; # has '
	    $score_r += 7 if $f eq "CCV";
	    $score_r += 8 if $f eq "CVV" and length($l[$_]) == 3; # no '
		
	    if ($f eq "CVC") {
		my $pair = substr($l[$_], -1, 1) . substr($l[$_+1],0,1);
		unless (grep {$_ eq $pair} @cpairs) {
		    # need a y-hyphen
		    print " needed y-hyphen\n"
		      if $make_lujvo_verbose;
		    $l[$_] .= "y";
		    $score_h++;
		}
	    }
	    if ($f eq "CVV") {
		# need r/n-hyphen after CVV rafsi:
		# 1 if following piece is a gismu (length==5)
		# 2 if the CVV is in the initial position of a 3+ part lujvo
		#   (else CVV could be split off as a cmavo and rest of lujvo
		#   is still valid lujvo: the hyphen insures that the rest
		#   won't be valid without the CVV bit)
		# 3 if an initial CVV is followed by another CVV (to insure
		#   a consonant cluster in the first 5 letters
		if (length($f[$_+1]) == 5   		# 1
		    or ($_ == 0 and @f > 2) 		# 2
		    or ($_ == 0 and $f[$_+1] eq "CVV"))	# 3
		{
		    # need it
		    $score_h++;
		    if (substr($l[$_+1],0,1) eq "r") {
			# n-hyphen
			$l[$_] .= "n";
		    } else {
			$l[$_] .= "r";
		    }
		}
	    }
	}
	# 4-letter rafsi get a "y" tacked on the end
	map { 
	    if (length($f[$_]) == 4) {
		$l[$_] .= "y";
		$score_h++;
	    }
	} (0 .. $#l);
	my $word = join('',@l);
	push(@lujvo, $word);
	# compute a score for good measure
	my $score_l = length($word);
	my $score_a = $word =~ tr/\'//;
	my $score_v = $word =~ tr/aeiou//;
	$scores{$word} =  1000 * $score_l - 500 * $score_a + 100 * $score_h
	  - 10 * $score_r - $score_v;
    }
    @lujvo = sort { $scores{$a} <=> $scores{$b} } @lujvo;
    return(@lujvo);
}

# return an array of possibilities, each an arrayref
sub make_lujvo_recurse {
    my($rafsis, $level) = @_;
    #print " recurse(level $level, rafsis ",scalar(@$rafsis),")\n";
    my(@results);
    if ($level < $#$rafsis) {
	#print " nonterminal\n";
	my(@sub) = make_lujvo_recurse($rafsis, $level+1);
	foreach my $choice (@{$rafsis->[$level]}) {
	    foreach (@sub) {
		my(@new) = @$_;
		unshift(@new, $choice);
		push(@results, \@new);
	    }
	}
	#print " returning ",scalar(@results)," possibilities\n";
	return(@results);
    } else {
	#print " terminal: returning ",scalar(@{$rafsis->[$level]})," possibilities\n";
	return(map {[$_]} @{$rafsis->[$level]});
    }
}

my $split_lujvo_debug = 1;

sub split_lujvo {
    my($lujvo) = @_;
    my(@g);
    my $form = categorize($lujvo);
  component:
    while(length($lujvo)) {
	if (length($lujvo) == 5 and length($form) == 5) {
	    push(@g, $lujvo);
	    substr($lujvo,0,5) = ""; substr($form,0,5) = "";
	    next component;
	}
	if (length($lujvo) >= 5 and substr($lujvo,4,1) eq 'y') {
	    # the first component is a 4rafsi
	    my $gismu = substr($lujvo,0,4);
	    substr($lujvo,0,5) = ""; substr($form,0,5) = "";
	    # find the right gismu by checking each vowel
	    foreach (qw(a e i o u)) {
		if (exists($rawgismu{$gismu . $_})) {
		    # found it
		    push(@g, $gismu . $_);
		    next component;
		}
	    }
	    if ($split_lujvo_debug) {
		print "4rafsi '$gismu' without matching gismu\n";
		print "so far: ",join('-',@g),", left: $lujvo, form left: $form\n";
		die;
	    }
	    return undef; # malformed
	}
	# should be a 3rafsi, possibly with a hyphen (y,r,n)?
	if (length($lujvo) < 3) {
	    if ($split_lujvo_debug) {
		print "remaining lujvo too short to be 3rafsi\n";
		print "so far: ",join('-',@g),", left: $lujvo, form left: $form\n";
		die;
	    }
	    return undef;
	}

	my $rafsi = substr($lujvo,0,3);
	if ($rafsi =~ /\'$/) {
	    $rafsi = substr($lujvo,0,4);
	}
	my $rafsiform = substr($form,0,3);
	my $g = rafsi2gismu($rafsi);
	unless ($g) {
	    # malformed
	    if ($split_lujvo_debug) {
		print "3rafsi '$rafsi' without matching gismu\n";
		print "so far: ",join('-',@g),", left: $lujvo, form left: $form\n";
		die;
	    }
	    return undef;
	}
	push(@g, $g);
	# now, is the next letter a hyphen?
	my $is_hyphen = 0;
	# a hyphen may have been neeed to insure a consonsant cluster in the
	# first 5 letters. This must be between rafsi (if the first term was a
	# 4rafsi, it would already have the cluster). It must be of the form
	# CVVrCVV
	if ($form =~ /^CVVCCVV/ and substr($lujvo,length($rafsi),1) =~ /[rn]/){
	    # the r or n is a hyphen
	    $is_hyphen = 1;
	}
	substr($lujvo,0,length($rafsi)) = ""; substr($form,0,3) = "";

	# the following component could be any of:
	#  r CVV (there are no rCV rafsi)
	#  r CCV (there are no rCC rafsi, in fact no CCC rafsi for that matter)
	#  r CVC (no rCV rafsi)
	#  r CCVCV (no CCC)
	#  r CVCCV (no rCV)
	#  rVV  rafsi
	#  rCV  none
	#  rVC  rafsi
	#  rCVCV only at end: length==5
	#  rVCCV only at end: length==5
	# if it's an r or n, and the next letter is a C, then the r or n was 
	# a hyphen
	unless($is_hyphen) {
	    if (length($lujvo) != 5 and $lujvo =~ /^[rn]/ and $form =~ /^CC/) {
		# is a hyphen
		$is_hyphen = 1;
	    }
	    # y hyphen? y is always a hyphen
	    if ($lujvo =~ /^y/) {
		$is_hyphen = 1;
	    }
	}
	if ($is_hyphen) {
	    substr($lujvo,0,1) = ""; substr($form,0,1) = "";
	}
	next component;
    }
    # we're done
    return(@g);
}


# Autoload methods go after =cut, and are processed by the autosplit program.

1;
__END__

=head1 NAME

Lingua::Lojban - Perl module for Lojban language processing

=head1 SYNOPSIS

  use Lingua::Lojban qw(gismu rafsi2gismu rafsi4gismu make_lujvo split_lujvo);

  $gref = gismu("karce"); # hashref of gismu properties
  $gismu = rafsi2gismu("fi'u"); # returns "cfipu"
  @rafsi = rafsi4gismu("condi"); # returns ("con", "cno", "coi")
  @lujvo = make_lujvo("xarju","birje","bilma");
  # returns list of all possible lujvo for those gismu, sorted so that the
  # best one comes first
  @gismu = split_lujvo("tankysi'erxarnu"); # returns qw(tanko snime xarnu)

  use Lingua::Lojban qw(gismulist rafsilist);
  @all_gismu = gismulist(); # all gismu
  @all_rafsi = rafsilist(); # all rafsi

=head1 DESCRIPTION

This module is meant to help a little bit in parsing and looking up Lojban
words, providing the basis for a flash card program, a glosser, or other
lookup programs.

The gismu and rafsi databased are stored in GDBM files which must be built for
your particular architecture (they are byte-order dependent, at least).  Two
scripts, "builddb.pl" and "buildrafsidb.pl" are provided for this purpose.
The "gismu.gdbm" and "rafsi.gdbm" data files are installed next to the
Lojban.pm file at installation time. Lojban.pm looks for them there or in the
current or parent directory (to aid while testing before installing). The
"gismu.fixed" and "rafsi.fixed" files are used to create these databases. The
gismu list is the same as the official gismu list from the Lojban FTP site but
with the 151 non-5-letter gismu removed. The rafsi list is similar: it only
contains rafsi that represent 5-letter gismu (no cmavo).

=head2 gismu

 $g = gismu("fagri");

Returns a hashref with all the various properties of the given gismu. The
keys of this hash are:

 gismu : the original gismu
 rafsi1,rafsi2,rafsi3 : the three rafsi that represent this gismu. each is
                        either a rafsi or undef.
 english : the one-word english equivalent
 metaphor : a description in english
 place_structure : the place structure for the gismu
 code : section/subsection for this word in an early draft textbook, gives
        a basic classification
 number : a frequency of use. Higher numbers == more frequently used
 refs : cross-references to related gismu

The gismu are looked up from the GDBM file, and are cached when loaded to
speed up access.

=head2 gismulist

 @all_gismu = gismulist();

Returns an array of all gismu, in random order.

=head2 rafsilist

 @all_rafsi = rafsilist();

Returns an array of all rafsi, in random order.

=head2 rafsi2gismu

 $gismu = rafsi2gismu("ger"); # returns "gerku"

Returns the gismu that the given rafsi represents. Returns undef if it was not
a rafsi for a normal 5 letter gismu (not a rafsi or was a rafsi for a cmavo).

=head2 rafsi4gismu

 @rafsi = rafsi4gismu("ger"); # returns ("ger","ge'u")

Returns the list of rafsi that represent the given gismu. From 0 to 3 are
returned. Undef is returned if the word is not a normal 5-letter gismu.

=head2 make_lujvo

=head2 split_lujvo


=head1 AUTHOR

Brian Warner, warner@lothar.com

=head1 SEE ALSO

The Lojban WWW archive: <http://xiron.pc.helsinki.fi/lojban/>
My home page for new versions: <http://www.lothar.com>

=cut

=notes

notes:

cmavo are in one of 4 forms: V (.a, .e, ...), CV, VV (.au ..), CVV (ki'a, pei)'
 (experimental cmavo are xVV or VVV* or CVVV*)
compound cmavo: (p. 51)
 have no consonant pairs or clusters
 end in a vowel
 to break up, first split(/\./)
  then for each component pull off from left V+:any contiguous vowels (V or VV)
   or CV+ (CV or CVV)
   ignoring "'"

cmeme:
 end in a consonant and a pause (C.)

brivla: (p. 53)
 end in a vowel
 have a consonant pair in the first five letters (ignoring "y")
 (have two or more syllables)
 have 3 subtypes: gismu, lujvo, fu'ivla    '
 
gismu: (p. 54)
 5 letters
 CVCCV or CCVCV (CCVCV requires an initial consonant cluster)
 contain no "'"

fu'ivla: (p. 61)    '
 stolen/borrowed/loan words
 liberal definition of what forms are legal (p. 62)



check for:
 CCV+CVC+something looking like CCVCV+...
 CVC+CVC+something looking like CVCCV+...

rafsi are 5, 4, or 3 :CVV,CCV, or CVC

 useful note:
there are no rafsi of the form rCV,nCV, or lCV. So CVC+([rn]+CVC), stripped of
initial CVC leaving [rn]CVC, cannot be misinterpreted as [rn]CV+C...

CVC cannot be at end (must end with a vowel)
4rafsi for all but end
5rafsi for end

if gismu, end 
substr($_,4,1) eq 'y', first part is 4rafsi

    #lorxu+rapli+barda+curnu (fox+repeat+big+worm)
    # lor y rap y bra curnu  == loryrapybra curnu ?
    # it seems that that would break up, however maybe it's ok because it would
    # have exactly the same meaning

=cut
