#!/usr/local/bin/perl -w

use GDBM_File;
use Getopt::Long;
use Storable qw(freeze thaw);

# hey.. I have a duplex printer at work. Cool.. do it in postscript to make
# it more difficult.

# ideally each card would have many sides:
#  gismu
#  rafsi * 3
#  english
#  place structure (basically includes english)

# put a shaded triangle in the top right corner of the front side of each card
# gismu goes centered on the top half of the front
# rafsi goes left, center, right on the front bottom
# english goes centered on the top back
# place structure is formatted on the bottom back
#  no, formatting is impossible, just wrap it
#  actually, wrapping is impossible too.

# that way it should be possible to expose just enough of the card to test the
# stuff I want to

# accept an argument that points to a file, in which each line starts with a
# gismu to be printed in this session. Accept arguments that indicate a
# numerical range of gismu to print (from that selection). Also take arguments
# to only print certain pages of the output, and only front or back sides (to
# test at home with a manually-fed spoofed-duplex printer)

sub moveto {
    my($x, $y) = @_;
    my(@ps);
    push(@ps, "$x 72 mul $y 72 mul moveto");
    return(@ps);
}
sub lineto {
    my($x, $y) = @_;
    my(@ps);
    push(@ps, "$x 72 mul $y 72 mul lineto");
    return(@ps);
}
sub centershow {
    my($x, $y, $string) = @_;
    $string =~ s/([\(\)])/\\$1/g;
    my(@ps);
    push(@ps, "$x 72 mul $y 72 mul moveto");
    push(@ps, "($string) dup stringwidth pop 2 div neg 0 rmoveto show");
    return(@ps);
}
sub show {
    my($string) = @_;
    $string =~ s/([\(\)])/\\$1/g;
    my(@ps);
    push(@ps, "($string) show");
    return(@ps);
}

use Font::Metrics::HelveticaBold;
use Font::Metrics::TimesRoman;
use Font::Metrics::NewCenturySchlbkRoman;

# hacked from version in Font::AFM.pm
sub stringwidth
{
    my($font, $string, $pointsize) = @_;
    return 0.0 unless defined $string;
    return 0.0 unless length $string;
    my @wx = @{"Font::Metrics::${font}::wx"};
    my $width = 0.0;
    for (unpack("C*", $string)) {
	$width += $wx[$_];
    }
    if ($pointsize) {
	$width *= $pointsize;
    }
    $width;
}

## @lines = pswrap($text, $font, $fontsize, $width_inches)

sub pswrap {
    my($t, $font, $size, $width) = @_;

    my(@lines,$line);

    while(1) {
	# search forward for the next word boundary. boundaries for this kind
	# of text are: space and the right side of a "/"
	$t =~ s/^\s+//;
	last unless length($t);
	$t =~ /^([^\/\s]+\/?)/;
	my $word = $1;
	die "no word, t:'$t',line:'$line'" unless defined($word);
	my $wordwidth = stringwidth($font, $word, $size);
	my $linewidth = stringwidth($font, $line, $size);
	if ($linewidth + $wordwidth > $width) {
	    # won't fit. Need a new line
	    if (!$linewidth) {
		# we're already on a new line. clip it at the line width
		$word = "";
		while(stringwidth($font,$word,$size) < $width) {
		    $t =~ s/^(.)//;
		    $word .= $1;
		}
		push(@lines, $word);
		$line = "";
		next;
	    }
	    push(@lines, $line);
	    $line = "";
	    next;
	} else {
	    # will fit. Add it to the current line. Tack on whitespace.
	    $t =~ s/^(\Q$word\E\s*)//;
	    $line .= $1;
	}
    }

    push(@lines, $line) if length($line);
    return(@lines);
}

# returns an array of postscript lines to draw one side of one card
sub do_card {
    my($gismu, $front, $card) = @_;
    my(@ps);
    print "$gismu card $card front $front\n";
    my($rafsi1, $rafsi2, $rafsi3, $english, $place) =
      @{$gismu{$gismu}}{qw(rafsi1 rafsi2 rafsi3 english place_structure)};
#    print " $english  :$rafsi1:$rafsi2:$rafsi3:\n";
#    my($pre_x1) = ($place =~ /^(.*)x1/);
#    print " pre: $pre_x1\n" if defined($pre_x1);
    # cards should be about 3x5, which means 3 across and 2 down on a 8.5x11
    # sheet. Print some tick marks for cutting alignment.
    # 012   x:0.5-3.0,3.0-5.5,5.5-8.0
    # 345   y:0.5-5.5,5.5-10.5
    my $xslot = $card % 3;
    $xslot = 2 - $xslot unless $front; # duplex left/right?
    my $yslot = $card > 2 ? 1 : 0;
    # remembering my postscript:
    # coordinates are x,y, origin is lower left, 72 per inch
    my $x = (0.5 + $xslot * 2.5);
    my $y = (0.5 + 4.5 - $yslot * 4.5);
 
    # draw the card border
    push(@ps, "$x 72 mul $y 72 mul moveto");
    push(@ps, "$x 2.5 add 72 mul $y 72 mul lineto");
    push(@ps, "$x 2.5 add 72 mul $y 4.5 add 72 mul lineto");
    push(@ps, "$x 72 mul $y 4.5 add 72 mul lineto");
    push(@ps, "$x 72 mul $y 72 mul lineto");
    push(@ps, "stroke");
    # mark the top-right corner
    if ($front) {
	push(@ps, moveto($x+2.25, $y+4.5));
	push(@ps, lineto($x+2.5, $y+4.25));
	push(@ps, "stroke");
    }

    if ($front) {
	# draw the gismu on top, the three rafsi (if present) on bottom
	# gismu is at 1.5 from the top (4.0 from bottom), 1.0 from left
	push(@ps, "/Helvetica-Bold 20 selectfont");
	push(@ps, centershow($x+1.25, $y+3.5, $gismu));
	if ($rafsi1) {
	    push(@ps, "/Times-Roman 15 selectfont");
	    push(@ps, centershow($x+0.5, $y+0.7, $rafsi1));
	}
	if ($rafsi2) {
	    push(@ps, "/Times-Roman 15 selectfont");
	    push(@ps, centershow($x+1.25, $y+0.5, $rafsi2));
	}
	if ($rafsi3) {
	    push(@ps, "/Times-Roman 15 selectfont");
	    push(@ps, centershow($x+2.0, $y+0.7, $rafsi3));
	}
    } else {
	# draw the english on top, place structure on bottom
	# have to wrap place structure?!?
	# need to clip everything to the current box, at least
	push(@ps, "/NewCenturySchlbk-Roman 18 selectfont");
	push(@ps, centershow($x+1.25, $y+3.5, $english));
	my(@lines) = pswrap($place, "NewCenturySchlbkRoman", 12, 2.1 * 72);
	push(@ps, "/NewCenturySchlbk-Roman 12 selectfont");
	foreach my $i (0 .. $#lines) {
	    push(@ps, moveto($x+0.2, $y+1.8 - $i*12/72));
	    push(@ps, show($lines[$i]));
	}
    }
    return(@ps);
}


tie %rawgismu, "GDBM_File", "gismu.gdbm", &GDBM_READER, 0644
  or die "couldn't read database: $!";
# The gismu file is indexed by gismu. Each value is a frozen hashref of:
#  gismu
#  rafsi1 (CVC), rafsi2(CCV), rafsi3(CV'V)
#  english: a single-word english equivalent
#  metaphor: how the gismu might be used as a metaphor
#  place_structure: x1 is foo by x2 with x3 etc
#  code: lesson/sublesson number
#  number: frequency count
#  refs: (cf. other gismu)
print "thawing gismu..\n";
foreach (keys(%rawgismu)) {
    $gismu{$_} = thaw($rawgismu{$_});
}
print " done\n";

@gismu = sort(keys(%gismu));
print "db has ",scalar(@gismu)," gismu\n";

# come up with a list of cards to produce this run.
@sorted = sort {$gismu{$a}->{'number'} <=> $gismu{$b}->{'number'}} @gismu;
#for $i (0 .. 10) {
#    my $gismu = $sorted[$i];
#    print "$i: (",$gismu{$gismu}->{'number'},") $gismu\n";
#}

#for $i (($#sorted - 10) .. $#sorted) {
#    my $gismu = $sorted[$i];
#    print "$i: (",$gismu{$gismu}->{'number'},") $gismu\n";
#}

$start = 251;
$end = 192; # total cards: $start - $end + 1
# done: 11..0, 12..71, 72..131, 132..191, 192..251

@cards = @sorted[($#sorted - $start) .. ($#sorted - $end)];
$do_fronts = 1;
$do_backs = 1;
print "printing ",scalar(@cards)," cards\n";

open(PS, ">foo2.ps") or die;
print PS '%!PS-Adobe-2.0' . "\n";
$pages = int($#cards / 6) + 1;
foreach $i (0 .. $#cards) {
    my $page = int($i / 6);
    push(@{$fronts[$page]}, do_card($cards[$i], 1, $i % 6)) if $do_fronts;
    push(@{$backs[$page]}, do_card($cards[$i], 0, $i % 6)) if $do_backs;
}
foreach (0 .. $pages-1) {
    if ($do_fronts) {
	print PS join("\n",@{$fronts[$_]}),"\n";
	print PS "showpage\n";
    }
    if ($do_backs) {
	print PS join("\n",@{$backs[$_]}),"\n";
	print PS "showpage\n";
    }
}
close(PS);
